/*
**
** Enumerating Hurdles
** -------------------
**
** Author: Dimitris Diochnos
**
** Date Started : Tuesday, 15th June 2004.
** Date Finished: Monday,  21st June 2004.
**
*/


#include <stdio.h>
#include <malloc.h>			/* Necessary for memory allocations. */
#include <math.h>			/* Necessary for ceil () function. */
#include <stdlib.h>			/* Necessary for atoi () function. */
#include <string.h>			/* Necessary for string manipulation. */
#include <time.h>			/* Necessary for time () function. */
#include "moret.h"			/* Necessary for Moret et al. linear time algorithm. */


/*******************************************************************/

#define OK 1
#define MAX_LINE_LENGTH	13	/* Max number of characters per line in input file. */

/*******************************************************************/

struct Permutation {
	int		Elements;
	int *	Array;
};

typedef struct Permutation PERMUTATION;
typedef PERMUTATION * PERMUTATIONPTR;

/*******************************************************************/

enum Algorithm { INVALID = 0, CUBIC, QUADRATIC, LINEAR };

/*******************************************************************/

void		Usage_Message (FILE *);
PERMUTATIONPTR	Permutation_Initialize (const int);
int		Permutation_Destroy (PERMUTATIONPTR *);
int		Set_Identity_Permutation (PERMUTATIONPTR);
int		Permutation_Print (PERMUTATIONPTR);
int	*	Count_Hurdles_N3 (PERMUTATIONPTR);
int	*	Count_Hurdles_N2 (PERMUTATIONPTR);
int *		Count_Hurdles_N (PERMUTATIONPTR);

/*******************************************************************/



int	main (int argc, char * argv [])
{
	PERMUTATIONPTR	permutation;
	register int	temp, size, num_hurdles, i, algo = INVALID;
	int *		hurdles;
	FILE *		input, * output, * time;
	char *		outputName, * timeName;
	char		line [MAX_LINE_LENGTH];
	clock_t 	start, finish, start_reading, finished_reading;
	clock_t 	start_calculating, finished_calculating;

	start = clock ();

	if (argc != 3) {
		Usage_Message (stderr);

		return	-1;		/* Indicate error. */
	}
	else if ((strcmp (argv [1], "linear") != 0) && (strcmp (argv [1], "quadratic") != 0) && (strcmp (argv [1], "cubic") != 0)) {
		Usage_Message (stderr);

		return	-2;		/* Indicate error. */
	}

	/* Open input file for reading. */
	input = fopen ( argv [2] , "r" );

	if (input == (FILE *) NULL) {
		fprintf (stderr, "File does not exist. Please retry with another file name.\n");

		return	-3;		/* Indicate error. */
	}

	/* File exists. Verify correct format. */
	temp = strlen (argv [2]);
	if ( (strcmp ( &(argv [2][temp - 3]), ".in" ) != 0 )  &&  (strcmp ( &(argv [2][temp - 3]), ".IN") != 0 ) ) {
		fprintf (stderr, "Incorrect input file name.\n");
		fprintf (stderr, "Please verify that input file name has extension \".in\"");

		fclose (input);
		return	-4;		/* Indicate error. */
	}

	/* Indicate reading input file. */
	fprintf (stdout, "Reading data from file. Please wait ...                             ");
	fflush (stdout);
	start_reading = clock ();

	/* Read data from input file - Assume correct format. */
	fgets (line, MAX_LINE_LENGTH, input);
	size = atoi (line);
	permutation = Permutation_Initialize (size - 2);
	i = 0;
	while (fgets (line, MAX_LINE_LENGTH, input) != (char *) NULL) {
		permutation->Array [i] = atoi (line);
		i++;
	}
	fprintf (stdout, "DONE!\n");
	finished_reading = clock ();

	fprintf (stdout, "Calculating hurdles using ");
	start_calculating = clock ();

	/* Calculate Hurdles. */
	if (strcmp (argv [1], "linear") == 0) {
		algo = LINEAR;
		fprintf (stdout, "linear-time algorithm. Please wait ...    ");
		fflush (stdout);
		hurdles = Count_Hurdles_N (permutation);
	}
	else if (strcmp (argv [1], "quadratic") == 0) {
		algo = QUADRATIC;
		fprintf (stdout, "quadratic-time algorithm. Please wait ... ");
		fflush (stdout);
		hurdles = Count_Hurdles_N2 (permutation);
	}
	else {
		algo = CUBIC;
		fprintf (stdout, "cubic-time algorithm. Please wait ...     ");
		fflush (stdout);
		hurdles = Count_Hurdles_N3 (permutation);
	}
	fprintf (stdout, "DONE!\n");
	finished_calculating = clock ();

	fprintf (stdout, "Generating files with results. Please wait ...                      ");
	fflush (stdout);

	/* Set Output and Time file names. */
	outputName = (char *) malloc (sizeof (char) * (temp + 2));
	timeName = (char *) malloc (sizeof (char) * (temp + 7));
	strncpy (&(outputName [0]), argv [2], sizeof (char) * (temp - 2) );
	strncpy (&(timeName [0]), argv [2], sizeof (char) * (temp - 3) );
	strncpy (&(timeName [temp - 3]), "_time.txt", sizeof (char) * 9);
	if ( (strcmp ( &(argv [2][temp - 3]), ".in" ) == 0 ))
		strncpy (&(outputName [temp - 2]), "out", sizeof (char) * 3);
	else
		strncpy (&(outputName [temp - 2]), "OUT", sizeof (char) * 3);
	outputName [temp + 1] = '\0';
	timeName [temp + 6] = '\0';

	/* Open Output and Time file descriptors. */
	output = fopen (outputName, "w");
	time = fopen (timeName, "w");
	if ((output == (FILE *) NULL) || (time == (FILE *) NULL)) {
		fprintf (stderr, "\nError generating %s or %s file.\n", outputName, timeName);
		fprintf (stderr, "Perhaps there is a conflict with an open file.\n");
		fprintf (stderr, "Please close all files and retry.\n");

		/* Close file descriptors. */
		fclose (input);
		if (output != (FILE *) NULL)
			fclose (output);
		if (time != (FILE *) NULL)
			fclose (time);

		/* Free-up memory. */
		Permutation_Destroy (&permutation);
		free (hurdles);
		free (outputName);
		free (timeName);

		return	-5;		/* Indicate error. */
	}

	/* Write data to Output file. */
	num_hurdles = 0;
	for (i = 0; i < permutation->Elements + 1; i++)
		if (hurdles [i] != 0)
			num_hurdles++;
	fprintf (output, "%d\n", num_hurdles);
	for (i = 0; i < permutation->Elements + 1; i++)
		if (hurdles [i] != 0)
			fprintf (output, "%d %d\n", i, i + hurdles [i] - 1);

	/* Write date to Time file. */
	fprintf (time, "Time Elapsed\n");
	fprintf (time, "------------\n");
	fprintf (time, "* Algorithm Implemented: ");
	switch (algo) {
		case CUBIC:
			fprintf (time, "CUBIC.\n");
			break;
		case QUADRATIC:
			fprintf (time, "QUADRATIC.\n");
			break;
		default:
			fprintf (time, "LINEAR.\n");
			break;
	}
	fprintf (time, "* Read data from file  : %9.3f msecs.\n", (double) 1000 * (finished_reading - start_reading) / CLOCKS_PER_SEC);
	fprintf (time, "* Calculate Hurdles    : %9.3f msecs.\n", (double) 1000 * (finished_calculating - start_calculating) / CLOCKS_PER_SEC);
	finish = clock ();
	fprintf (time, "* File Generation      : %9.3f msecs.\n", (double) 1000 * (finish - finished_calculating) / CLOCKS_PER_SEC);
	fprintf (time, "* Entire Operation     : %9.3f msecs.\n", (double) 1000 * (finish - start) / CLOCKS_PER_SEC);

	/* Close file descriptors. */
	fclose (input);
	fclose (output);
	fclose (time);

	/* Indicate successful completion. */
	fprintf (stdout, "DONE!\n\n");
	fprintf (stdout, "Time elapsed: %9.3f msecs.\n\n", (double) 1000 * (finish - start) / CLOCKS_PER_SEC);
	fprintf (stdout, "Refer to:\n");
	fprintf (stdout, "* %s     : Requested output.\n", outputName);
	fprintf (stdout, "* %s: Time information for various operations.\n", timeName);

	/* Free-up memory. */
	Permutation_Destroy (&permutation);
	free (hurdles);
	free (outputName);
	free (timeName);


	return	0;	/* Successful completion. */
}


/*******************************************************************/

void			Usage_Message (FILE * stream)
{
	fprintf (stream, "Usage: hurdles <Algorithm> <InputFile>\n");
	fprintf (stream, "Where <Algorithm> = linear, quadratic, cubic\n");
	fprintf (stream, "Where <InputFile> = The file name containing the correct input.\n");

	return;
}


/*******************************************************************/

PERMUTATIONPTR	Permutation_Initialize (const int Elements)
{
	PERMUTATIONPTR	newPermutation;

	newPermutation = (PERMUTATIONPTR) malloc (sizeof (PERMUTATION));
	if (newPermutation == (PERMUTATIONPTR) NULL)
		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */

	newPermutation->Elements = Elements;
	/* Reserve extra storage for 0 and EOP. */
	newPermutation->Array = (int *) malloc (sizeof (int) * (newPermutation->Elements + 2));
	if (newPermutation->Array == (int *) NULL) {
		free (newPermutation);

		return	(PERMUTATIONPTR) NULL;	/* Indicate error. */
	}

	/* Set limits. */
	newPermutation->Array [0] = (int) 0;
	newPermutation->Array [newPermutation->Elements + 1] = (int) (newPermutation->Elements + 1);

	return	newPermutation;	/* Everything is OK. */
}

/*******************************************************************/

int				Permutation_Destroy (PERMUTATIONPTR * permutation)
{
	if ((*permutation) != (PERMUTATIONPTR) NULL) {
		free ((*permutation)->Array);
		free ((*permutation));
		*permutation = (PERMUTATIONPTR) NULL;
	}

	return	OK;		/* Memory is free! */
}

/*******************************************************************/

int * 			Count_Hurdles_N3 (PERMUTATIONPTR permutation)
{
	register int 	i, j, length, distance, k, frame, N;
	int * 		framed = (int *) malloc (sizeof (int) * (permutation->Elements + 2));

	N = permutation->Elements;

	for (i = 0; i < N + 2; i++)
		framed [i] = 0; 	/* Initialize to ZERO length. */

	/* Perform Check for Hurdles. */
	for (i = 0; i < N + 1; i++) {
		for (length = N + 2 - i; length >= 2; length--) {
			/*
			** Remember that minimum length is 4.
			** You could have started with length = 4 the for () loop.
			** Or finish repeat for () loop while length >= 4.
			** We start from (N + 2) so as to add a larger constant in O(n^3).
			*/
			j = (i + length - 1); 	/* Set j's position. */
			if (permutation->Array [j] == permutation->Array [i] + length - 1 ) {
				/* This is a candidate framed interval. */
				frame = 1;
				/* Check that all elements are in-bounds. */
				for (distance = 1; distance < length - 1; distance++) {
					k = i + distance; 	/* Set k's position. */
					if ((permutation->Array [k] < permutation->Array [i]) || (permutation->Array [k] > permutation->Array [j])) {
						/* This is out of bounds. */
						frame = 0;
						break; 	/* Quick exit. */
					}
				}

				if (frame) {
					/* Eliminate larger intervals. */
					framed [i] = length; 	/* Set number of elements this FI contains. */
				}
			}
		}
	}

	/* Now kill framed intervals that contain shorter framed intervals. */
	for (i = 0; i < N + 1; i++) {
		if (framed [i] != 0) {
			for (length = 2; length < framed [i]; length++) {
				j = i + length - 1; 	/* Set j's position. */
				if (framed [j] != 0) {
					framed [i] = 0;
					break; 	/* Only one is enough. */
				}
			}
		}
	}

	return 	framed;
}

/*******************************************************************/

int	*			Count_Hurdles_N2 (PERMUTATIONPTR permutation)
{
	register int	i, k, max, N, finish;
	register int	start;
	int *			framed = (int *) malloc (sizeof (int) * (permutation->Elements + 2));

	N = permutation->Elements;

	for (i = 0; i < N + 2; i++)
		framed [i] = 0;	/* Initialize to NULL. */

	/* Check for non-cyclic hurdles. */
	for (i = 0; i < N + 1; i++) {
		max = permutation->Array [i];
		k = 0;
		while (++k < N + 2 - i) {
			if (permutation->Array [i + k] < permutation->Array [i]) {
				/* Quick exit. */
				break;
			}
			else if ( (permutation->Array [i + k] == permutation->Array [i] + k) && (permutation->Array [i + k] == (max + 1)) ) {
				/* This IS a framed interval. */
				framed [i] = k + 1;
				break;
			}
			else if (permutation->Array [i + k] > max)
				max = permutation->Array [i + k];
		}
	}

	/* Now kill any framed intervals that include shorter framed intervals - Linear time. */
	i = 0;
	while (i < N + 1) {
		if (framed [i] != 0) {
			k = i;
			while (++k < framed [i] + i - 1) {
				/* Check for INCLUDED framed intervals. */
				if ( framed [k] != 0 ) {
					framed [i] = 0; 	/* This candidate was not a hurdle. */
					break;
				}
			}
			/*
			** Either this candidate should be now active,
			** or the end of the hurdle has been reached successfully.
			*/
			i = k - 1;
		}
		i++;
	}

	return	framed;
}


/*******************************************************************\
*********************************************************************
*********************************************************************
*********************************************************************
\*******************************************************************/

/***********************************\
* The following two functions stem  *
* from the linear time algorithm    *
* proposed by Moret et al. for      *
* calculating inversion distance.   *
\***********************************/


int *				Count_Hurdles_N (PERMUTATIONPTR permutation)
{
	int				cIdx, ind, i, j, N, j1, j2, size, num_hurdles, k;
	int *			oriented, *cc, *labeled;
	component_t *	components;
	int				num_components, num_oriented, first_comp, last_comp, num_block;
	int *			greyEdges, *perm;
	int *			invperm, *done, *cycle, next, c;
	distmem_t		distmem;
	int				great_s, great_f;
	int *			framed;

	N = permutation->Elements;
	size = 2 * N + 2;

	/* Memory allocation. */
	distmem.perm = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.perm == (int *)NULL)
		fprintf(stderr,"ERROR: perm NULL\n");
	distmem.done = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.done == (int *)NULL)
		fprintf(stderr,"ERROR: done NULL\n");
	distmem.greyEdges = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.greyEdges == (int *)NULL)
		fprintf(stderr,"ERROR: greyEdges NULL\n");
	distmem.stack = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.stack == (int *)NULL)
		fprintf(stderr,"ERROR: stack NULL\n");
	distmem.oriented = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.oriented == (int *)NULL)
		fprintf(stderr,"ERROR: oriented NULL\n");
	distmem.cc = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.cc == (int *)NULL)
		fprintf(stderr,"ERROR: cc NULL\n");
	distmem.labeled = (int *)malloc((2*N+2)*sizeof(int));
	if (distmem.labeled == (int *)NULL)
		fprintf(stderr,"ERROR: labeled NULL\n");
	distmem.components = (component_t *) malloc((2*N+2)*sizeof(component_t));
	if (distmem.components == (component_t *)NULL)
		fprintf(stderr,"ERROR: components NULL\n");
	/* end of memory allocation */

	/**********************\
	* STARTING num_cycles. *
	\**********************/

	done       = distmem.done;
	greyEdges  = distmem.greyEdges;
	invperm    = distmem.done;
	cycle      = distmem.labeled;

	perm = distmem.perm;
	perm [0] = 0;
	for (i = 1; i < N + 1; i++) {
		perm [2 * i - 1] = 2 * permutation->Array [i] - 1;
		perm [2 * i    ] = 2 * permutation->Array [i];
	}
	perm [2 * N + 1] = 2 * N + 1;

	/* set grey -edges */
	for (i=0 ; i<size ; i++) {
		invperm[perm[i]] = i;
		greyEdges[i] = -1;
	}

	j1 = invperm[1];
	if (j1 != 1)
		greyEdges[0] = j1;

	for (i=1 ; i<size-1 ; i+=2) {
		ind = perm[i];
		if (ind < perm[i+1]) {
		  j1 = invperm[ind-1];
		  j2 = invperm[ind+2];
		}
		else {
		  j1 = invperm[ind+1];
		  j2 = invperm[ind-2];
		}
		if (j1 != i-1)
		  greyEdges[i] = j1;
		if (j2 != i+2)
		  greyEdges[i+1] = j2;
	}

	j1 = invperm[size-2];
	if (j1 != size-2)
		greyEdges[size-1] = j1;

	for (i = 0 ; i < size ; i++)
		done[i] = 0;

	c = 0;
	for (i = 0 ; i < size ; i++) {
		if (done [i] == 0 && greyEdges [i] != -1) {
			cycle [i] = i;
			done [i] = 1;
			next = i;
			do {
				if (next % 2 == 0)
					next++;
				else
					next--;
				done [next] = 1;
				cycle [next] = i;
				next = greyEdges [next];
				done [next] = 1;
				cycle [next] = i;
			}
			while(next != i);
			c++;
		}
	}

	/***********************\
	* FINISHING num_cycles. *
	\***********************/

	/* By default, set number of hurdles and fortresses to 0 */
	num_hurdles = 0;

	/* Initializations. */
	oriented   = distmem.oriented;
	cc         = distmem.cc;
	labeled    = distmem.labeled;
	components = distmem.components;

	connected_component(size, &distmem, &num_components);

	/* Calculate if each gray edge is oriented or unoriented */
	/* At the same time, label the connected component of a vertex as the
	 index of its root */

	for (i=0 ; i<size ; i++) {
		j = greyEdges[i];
		if (j == -1)
		  oriented[i] = FALSE;
		else {
			if (i < j) {
				if ((j - i) % 2 != 0) {
					oriented[i] = FALSE;
					oriented[j] = FALSE;
				}
				else {
					oriented[i] = TRUE;
					oriented[j] = TRUE;
				}
			}
		}
	}

	/* Look for any vertices that are "labeled".
	 These are the roots of the cnnected components.
	 Record them in the "components" array,
	 and set num_components.
	*/

	/* If a component contains an oriented vertex then it is oriented.
	 Otherwise, the component is unoriented. */

	for (i=0 ; i<num_components ; i++)
		components[i].oriented = FALSE;

	for (i=0 ; i<size ; i++)
		if (oriented[i] == 1)
			components[cc[i]].oriented = TRUE;


	/* Count nonoriented components */
	num_oriented = 0;
	for (i=0 ; i<num_components ; i++)
		if (components[i].oriented == FALSE)
		  num_oriented++;

	for (i=0 ; i<num_components ; i++) {
		components[i].blocks   = 0;
		components[i].hurdle   = 0;
		components[i].left     = -1;
		components[i].right    = -1;
	}

	/* HURDLES
		Hurdles are a subset of the nonoriented components.
		There are two types of hurdles (in the KST-sense):
		"simple" and "superhurdle".
		First, we implicitly eliminate oriented components.
		Second, if a nonoriented component is one contiguous block of
		vertices, it is a hurdle.
		Third, if a hurdle "protects" the same non-hurdle on its left and
		right side, then it is a "superhurdle".
	*/

	first_comp = -1;
	last_comp = -1;
	num_block = -1;
	for (i=0 ; i<size ; i++) {
		cIdx = cc[i];
		if (cIdx != -1) {
			if (components[cIdx].oriented == FALSE) {
				if (cIdx != last_comp) {
					if (last_comp == -1)
						first_comp = cIdx;
					else {
						components[last_comp].right = cIdx;
						components[cIdx].left       = last_comp;
					}
					last_comp = cIdx;
					num_block++;
					components[cIdx].blocks++;
				}
			}
		}
	}


	for (i=0 ; i<num_components ; i++) {
		if ((components[i].oriented == FALSE) && (components[i].blocks == 1)) {
			components[i].hurdle = HURDLE;
			num_hurdles++;
		}
	}

	if ((first_comp == last_comp) && (components[first_comp].blocks == 2)) {
		components[first_comp].hurdle = (HURDLE | GREATHURDLE);
		num_hurdles++;
	}

	framed = (int *) malloc ( sizeof (int) * (N + 2));

	for (i = 0; i < N + 2; i++)
		framed [i] = 0;		/* Initialize to NULL. */

	i = 0;
	if ((first_comp == last_comp) && (components[first_comp].blocks == 2) && (num_hurdles > 1)) {
		while (i < size) {
			if ((components [cc [i]].blocks != 2) || (components [cc [i]].hurdle != (HURDLE | GREATHURDLE)))
				break;
			i++;
		}
		great_f = (int) ceil (i / 2);
	}

	while (i < size) {
		if ((components [cc [i]].blocks == 1) && (components [cc [i]].hurdle == HURDLE)) {
			k = (int) ceil (i / 2);
			i++;
			j = cc [i];
			while ((cc [i] == j) && (i < size))
				i++;
			framed [k] = ((int) ceil (i / 2)) - k + 1;		/* Set length. */
		}
		else if ((components [cc [i]].blocks == 2) && (components [cc [i]].hurdle == (HURDLE | GREATHURDLE)) ) {
			great_s = (int) ceil (i / 2);
			/* Do not count cyclic hurdle. */
			//framed [great_s] = great_f + N + 2 - great_s + 1;	/* Set length for great hurdle. */
			break;	/* Quick exit. */
		}
		else
			i++;
	}

	/* Free-up memory. */
	free (distmem.perm);
	free (distmem.done);
	free (distmem.greyEdges);
	free (distmem.stack);
	free (distmem.oriented);
	free (distmem.cc);
	free (distmem.labeled);
	free (distmem.components);

  return	framed;
}


/*******************************************************************/

void connected_component(int size,distmem_t *distmem, int *num_components)
{
  int i;
  int stack_ptr;
  int *greyEdges;
  int *cycle,*range,*cc;
  component_t *components;
  int *stack_root,*stack_range;
  int *parent,*next;
  int right,p;


  greyEdges            = distmem->greyEdges;

  next = stack_root    = distmem->stack;
  range = distmem->oriented;
  stack_range = cc     = distmem->cc;
  parent=cycle= distmem->labeled;
  components           = distmem->components;
  stack_ptr = -1;
  *num_components = 0;

  /* Use Linear algorithm to compute connected component */
  for (i=0; i<size ; i++) {
    if (greyEdges[i]==-1)
      continue;
    range[cycle[i]]=i;
  }

  for (i=0 ; i<size ; i++) {
    if (greyEdges[i]==-1)  continue;  /*it is self loop,discard it*/
    if (parent[i]==i) {
      stack_ptr++;
      stack_root[stack_ptr]=i;
      stack_range[stack_ptr]=range[i];
    }
    else { /*check the top of stack for intersection*/
      right=i;
      while (stack_root[stack_ptr] >parent[i]) {
        /*union top to the i's connected component*/
        parent[stack_root[stack_ptr]]=parent[i];
        if (right < stack_range[stack_ptr])
          right=stack_range[stack_ptr]; /*extend the active range*/
        stack_ptr--;
      }
      if (stack_range[stack_ptr] < right)
        stack_range[stack_ptr]=right;
      if (stack_range[stack_ptr] <=i) {
        /*the top connected-component is INACTIVE*/
        components[*num_components].index=stack_root[stack_ptr];
        (*num_components)++;
        stack_ptr--;
      }
    }
  }
  /*turn the forest to set of linked lists whose list head is index of
    component*/
  for (i=0 ; i<size ; i++)
    next[i]=-1;
  for (i=0 ; i<size ; i++) {
    if (greyEdges[i]==-1)
      cc[i]=-1;
    else
      if (i!=parent[i]) {
        /* insert i between parent(i) and next of parent(i) */
        next[i]=next[parent[i]];
        next[parent[i]]=i;
      }
  }

  /*label each node with its root*/
  for (i=0;i<*num_components;i++) {
    p=components[i].index;
    while(p!=-1) {
      cc[p]=i;p=next[p];
    }
  }

  return;
}

/*******************************************************************/
